/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#ifndef ALGUTILS_H
#define ALGUTILS_H

#include "postgres.h"
#include "utils/palloc.h"
#include "algstructs.h"

/* constants to help in the decision if to use an index scan when extending */
#define INDEX_AND_PLAN_PER_TUPLE_FACTOR 0.1
#define INDEX_AND_PLAN_PER_TUPLE_FACTOR_CPU_COST (0.001)
#define INDEX_AND_PLAN_BASE_FACTOR 0.1

/* 
 * adaptive upper bound algorithm constants. These are estimates for a pentium 1.6 machine
 * which is assumed to be a starting point. Perhaps in a few years these could be increased
 * a bit. but these are just a starting point, the algorithm increase them in mid runs anyway.
 * */
#define ADAPTIVE_EXTEND_UPPER_BOUND_MAX 500
#define ADAPTIVE_EXTEND_UPPER_BOUND_FIXED_PENALTY_INC 25
#define ADAPTIVE_EXTEND_UPPER_BOUND_FIXED_PENALTY_DEC 15

 /* note! the numbers matters! only add codes after last code! */
enum policies_update_codes
{
	POLICY_CODE_DELETED_NODE_0 = 0,
	POLICY_CODE_DELETED_NODE_1 = 1,
	POLICY_CODE_DELETED_NODE_2 = 2,
	POLICY_CODE_DELETED_NODE_3 = 3,
	POLICY_CODE_DELETED_NODE_4 = 4,
	POLICY_CODE_DELETED_NODE_5 = 5,
	POLICY_CODE_MAXIMALLY_EXTENDED_0 = 6,
	POLICY_CODE_MAXIMALLY_EXTENDED_1 = 7,
	POLICY_CODE_MAXIMALLY_EXTENDED_2 = 8,
	POLICY_CODE_MAXIMALLY_EXTENDED_3 = 9,
	POLICY_CODE_MAXIMALLY_EXTENDED_4 = 10,
	POLICY_CODE_MAXIMALLY_EXTENDED_5 = 11,
	POLICY_CODE_MAXIMALLY_EXTENDED_6 = 12,
	POLICY_CODE_NOT_INVOLVED_1 = 13,
	POLICY_CODE_NOT_INVOLVED_2 = 14,
	POLICY_CODE_NOT_INVOLVED_3 = 15
};

/* policy functions interface definition. Use this interface for new policy
 * functions.*/
typedef int (*POLICY_FUNC) (alg_fctx * fctx, memQueue * Q, int tsetsNum);

/* code that symbolises each policy */
#define POLICY_MAJORITY 0
#define POLICY_MINCOST 1
#define POLICY_RANDOM 2
#define POLICY_NAIVE 3

/* calculates a heaptuple size */
#define tupleSize(tuple)\
(HEAPTUPLESIZE + tuple->t_len)

/*
 * Call the type specific '=' function\
  */
#define areAttributesEqual_RT(fctx, attID, lvalue, rvalue)\
DatumGetBool(\
	   FunctionCall2(&(fctx->tupleSetAttEQFunctions[attID]->eq_opr_finfo) \
,lvalue, rvalue))

 /*
  * stpcpy is not mine but its difficult to get its header working all the
  * time. Thus i am redeclaring it here for safety. No harm anycase.
  */
extern char *stpcpy(char *__dest, const char *__src);

extern Datum myDatumCopy(Datum value, bool typByVal, int typLen);
extern void myDatumFree(Datum value, bool typByVal, int typLen);
extern void sortOidArray(Oid *array, int numElements);
extern void updateMaxTupleSize(alg_fctx * fctx, int relID, HeapTuple tuple);
extern void updateResultTupleMaxSize(alg_fctx * fctx);
extern TypeCacheEntry *getEQFunction(Oid typeid);
extern char *addAndExpandStringBy(char *oldStr, int *oldStrAllocSize, char *addStr
					 ,int sizeToAdd, char **relative_ptr, bool *free_old_str);
extern bool isResultTupleAttOccursInRelID(alg_fctx * fctx, int resultTupleAttID, int relID);
extern bool isResultTupleAttOccursInSomeRelID(alg_fctx * fctx, int resultTupleAttID
								  ,bytea *relsID);
extern bool isResultTupleAttOccursInSomeRelIDWithSpecifiedRelations(alg_fctx * fctx,
							   int resultTupleAttID, RBITS relationsToCheck);
extern bool isPlanUsingSeqScan(Plan *planTree);
extern void Extend(alg_fctx * fctx, memQueue * Q);
extern void updateOnDemandRelationsToScan(alg_fctx * fctx, memQueue * Q);
extern void onDemandPreCompleteRemovePage(alg_fctx * fctx, FuncCallContext *funcctx, memQueue * Q);
extern int	getRandomNextRelationToScan(alg_fctx * fctx, int lastScannedRelation);
extern void onDemandRetrieveNextPageFromR(alg_fctx * fctx, FuncCallContext *funcctx);
extern void OnDemandGetAlternativeTupleSets(alg_fctx * fctx, FuncCallContext *funcctx,
		  memQueue * out, bool trueUseMemorySizeFalseUseNumTupleSets, int SZ,
								int numTupleSetsToReturn, bool addToComplete);
extern int	algFloor(float value);
extern int	algCeiling(float value);
extern int	getAdaptiveUpperBound(alg_fctx * fctx);
extern void updateAdaptiveScore(alg_fctx * fctx, double time);

#endif   /* ALGUTILS_H */
