/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#ifndef QUEUES_H
#define QUEUES_H

#include "postgres.h"
#include "tset.h"


#define INFINIT INT_MAX

/* definition of a bit array that represents relations */
#include "c.h"

/* 
 * not all 64 bits systems or even 32 with limited 64 bits support can support
 * 64 bits bitwise operations!!!!! 
 * */
#ifdef BITS_64_BITWISE_SUPPORT
typedef uint64 RBITS;
#else
typedef uint32 RBITS;
#endif

/* memory queue node structure */
typedef struct
{
    /* tuple set */
	TSet		tset;
    /* 1 in each bit that repesent a relation we have visited */
	RBITS visited;
    /* 1 for each relation in the tuple set */
	RBITS relsOfTuples;
    /* 1 for each relation we potentially need to scan in the extending process */
	RBITS toVisit;
	struct memQueueNode *next;
}	memQueueNode;

/* memory queue */
typedef struct
{
	memQueueNode *head;
	memQueueNode *tail;
	int			numElements;
    /* actual bytes occupied in the queue */
	int			sizeInBytes;
    /* an upper bound memory allocated for this queue */
	int			allocatedBytes;
}	memQueue;


/* a storage queue that uses a tuple store */
typedef struct
{
    /* tuple store structure */
	Tuplestorestate *tuplestorage;
    /*number of elements in the queue */
	int			numElements;
    /* last position in the storage queue */
	int			lastReadPosition;
	int			sizeInBytes;
    /* marking a position in the queue */
	int			markedLastReadPosition;
	int			markedNumElements;
	int			markedSizeInBytes;
	HeapTuple	markedMemOverflowTuple;

	/*
	 * we pulled it out of storage but couldn't hold it memory. thus we leave
	 * it here. potentially i can read half tuples and thus never overflow,
	 * however, since i don't want to implement the tuplestore again, i
	 * simulate this by leaving this tuple in the storage structure. the next
	 * ead will be this tuple. in case of an overflow we do not increment the
	 * lastReadPosition.
	 */
	HeapTuple	memOverflowTuple;
    /* postgresql 8.2 compatibility */
#ifdef PG_MODULE_MAGIC
	TupleTableSlot *slot;
#endif   /* */
}	storageQueue;

#endif   /* QUEUES_H */
