/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */

#ifndef TSET_H
#define TSET_H

#include "postgres.h"
#include "utils/palloc.h"
#include "funcapi.h"

/* the position of the lables attribute in a tuple set. */
#define LABELS 0
/* sometimes the position of the attribute needs to be aligned to start
 * from 1. mainly for SPI and backend functions */
#define LABELS_ALIGNED 1
/* first attribute that is not a labels attribute in a tuple set*/
#define FIRSTATT 1
/* the last attribute position in a tuple set */
#define LASTATT (fctx->tupleSetNatt-1)
/* the number of attribute sin a tuple set */
#define NUMATT (fctx->tupleSetNatt)
/* the size of a CTID */
#define CTID_SIZE sizeof(ItemPointerData)
/* the size of a ctid and a table id together. this pair is actually 
 * a one label that represents a tuple in a tuple set */
#define TABLEID_CTID_SIZE (1+CTID_SIZE)

/* a deformed tuple set structure */
typedef struct
{
	int			size;
	/* tuple ids. TID or TID AND ROWID. 
     * Comprised of the id of the relation in our sorted oid array.
     * (not the oid itself) and the CTID of the tuple. and each pair
     * represent a tuple and many pairs represents the tuples in the
     * tuple set.*/
	bytea	   *tids;
    /* the heaptuple that represents the tuple set */
	HeapTuple	t;
	/* values */
	Datum	   *v;
	/* nulls */
	char	   *n;
}	deformedTSet;
typedef deformedTSet *TSet;

#define TSET_GET_CTID(tset,relative_pos) VARDATA(tset->tids)+relative_pos*(TABLEID_CTID_SIZE)+1
#define DTupleTupleHeaderSize 1

/* structure that represents a deformed tuple */
typedef struct
{
	HeapTuple	t;
	/* values */
	Datum	   *v;
	/* nulls */
	char	   *n;
    /* the id of the relation this tuple was taken from.
     * The id in the sorted relsOid and not the oid itself */
	ItemPointer tid;
    /* if true, the tuple is fully deformed */
	bool		deformed;
}	deformedTuple;
typedef deformedTuple *DTuple;

typedef struct
{
	HeapTupleHeader tup;
	Form_pg_attribute *att;
	int			tdesc_natts;
	/* number of atts to extract */
	int			natts;
	int			attnum;
	/* ptr to tuple data */
	char	   *tp;
	/* offset in tuple data */
	long		off;
	/* ptr to null bitmask in tuple */
	bits8	   *bp;
	/* can we use/set attcacheoff? */
	bool		slow;
    /* are we before or after reading the first attribute */
	bool		firstIteration;
}	deformTupleIterativeState;

#endif   /* TSET_H */
