/* 
 * Copyright (c) 2006 Itzhak Fadida.  All Rights Reserved. 
 * See the README file for a detailed BSD License.
 * This file and all related files are under the BSD license.
 * */
#ifndef TSETFUNCS_H
#define TSETFUNCS_H

#include "tset.h"
#include "algstructs.h"
#include "utils/palloc.h"

/* number of tuples in a tuple set. b is the labels attribute */
#define numOfTuplesInTSet(fctx, b)\
((VARSIZE(b) - VARHDRSZ) / (1 + sizeof(ItemPointerData)))

/* get the table id of the position-th tuple in the lables attribute b. */
#define getTableIdOfRelsOidInTSet(fctx, b, position)\
((int) VARDATA(b)[position * (1 + sizeof(ItemPointerData))])

/* this make sure that the tuple set it a heaptuple in memory and not deformed */
#define makeSureFormedTSet(fctx,tset)\
do\
{\
	if ((tset)->t == NULL)\
	{\
			(tset)->t = heap_formtuple((fctx)->tupleSetDesc, (tset)->v, (tset)->n);	\
			freeDatumArray(fctx, (tset)->v);\
			pfree((tset)->v);\
			pfree((tset)->n);\
			(tset)->v = NULL;\
			(tset)->n = NULL;\
		{\
				bool		isnull;	\
				(tset)->tids = (bytea *) fastgetattr((tset->t), LABELS_ALIGNED, (fctx)->tupleSetDesc, &isnull);\
		} \
			(tset)->size = tupleSize((tset)->t);\
	} \
} while (0)

/* this makes sure that the tuple set is also deformed into its v and n arrays. */
#define makeSureDeformedTSet(fctx,tset)\
do\
{\
		if ((tset)->v == NULL)\
	{\
			(tset)->v = (Datum *) palloc(NUMATT * sizeof(Datum));\
			(tset)->n = (char *) palloc(NUMATT * sizeof(char));	\
			heap_deformtuple((tset->t), (fctx)->tupleSetDesc, (tset)->v, (tset)->n);\
	} \
} while (0)

/* this makes sure that a heaptuple is also deformed into its v and n arrays */
#define makeSureDeformedTuple(fctx,tuple,tupleDesc,relID)\
do\
{\
		if ((tuple)->deformed == false)\
	{\
			heap_deformtuple((tuple->t), (tupleDesc), (tuple)->v, (tuple)->n);\
			tuple->tid = (ItemPointer) tuple->v[fctx->relsNAtt[relID]];\
			(tuple)->deformed = true;\
	} \
} while (0)

/* this frees an array of datums */
#define freeDatumArray(fctx,arr)\
do\
{\
		int			i;\
		for (i = 0; i < NUMATT; i++)\
			if ((arr)[i] != (Datum) NULL)\
				myDatumFree((arr)[i], (fctx)->tupleSetDesc->attrs[i]->attbyval, \
							(fctx)->tupleSetDesc->attrs[i]->attlen);\
} while (0)

/* this sets a heaptuple tuple set into its TSet structure */
#define setTSet(fctx,tset,tupleTSet,isnull)\
do\
{\
	(tset)->size = tupleSize((tupleTSet));\
		(tset)->t = (tupleTSet);\
		(tset)->tids = (bytea *) fastgetattr((tupleTSet), LABELS_ALIGNED, (fctx)->tupleSetDesc, &(isnull));\
} while (0)

extern bool tsetAttIsNull(TSet tset, int pos);
extern void freeTSet(alg_fctx * fctx, TSet tset);
extern TSet newEmptyTSet(alg_fctx * fctx);
extern Datum setTSetValue(alg_fctx * fctx, TSet tset, Datum value
			 ,int position);
extern Datum setTSetValueNoCopy(alg_fctx * fctx, TSet tset, Datum value, int position);
extern TSet newTSet_(alg_fctx * fctx);
extern TSet newTSet(alg_fctx * fctx, HeapTuple tset);
extern DTuple newDTuple(alg_fctx * fctx, int numAtt);
extern TSet newTSet___(alg_fctx * fctx, HeapTuple tset);
extern TSet newTSet____(alg_fctx * fctx);
extern char *printTSetToCStr(alg_fctx * fctx, bytea *b);
extern void printTSet(alg_fctx * fctx, HeapTuple tset);
extern void printDTSet(alg_fctx * fctx, TSet tset);
extern void printDTuple(alg_fctx * fctx, DTuple dtuple, int tupleRelID);
extern bytea *createNewTSetRowIDs(alg_fctx * fctx, ItemPointer tid, int tupleInRelsOid);
extern bytea *addRowIDtoTSet__(alg_fctx * fctx, bytea *baseSet, bytea *setContainingTRID
				 ,int tupleInRelsOid);
extern bytea *addRowIDtoTSet_(alg_fctx * fctx, bytea *b, ItemPointer tid, int tupleInRelsOid);
extern bytea *addRowIDtoTSet(alg_fctx * fctx, TSet tset, ItemPointer tid, int tupleInRelsOid);
extern bytea *mergeRowIDs(alg_fctx * fctx, bytea *bLeft, bytea *bRight);
extern bool areTSetsEqual(alg_fctx * fctx, TSet tset1, TSet tset2);
extern bool isLeftSubsumedRightTset(alg_fctx * fctx, TSet left, TSet right);
extern TSet JCCTSetAndTupleAndReturnTSet(alg_fctx * fctx, TSet tset, DTuple tuple
							 ,TupleDesc tupleTupleDesc, int tupleRelID, deformTupleIterativeState * dsTuple);
extern TSet unifyAndJCC(alg_fctx * fctx, TSet leftTSet, Datum *leftTSetVals,
			char *leftTSetNulls, TSet rightTSet, RBITS relationsToCheck,
			RBITS relationsOfLeft, RBITS relationsOfRight);
extern TSet simpleUnify(alg_fctx * fctx, TSet leftTSet, Datum *leftTSetVals,
			char *leftTSetNulls, TSet rightTSet);
extern TSet create_tuple_set(alg_fctx * fctx, DTuple inBaseTuple
				 ,TupleDesc baseTupleDesc, int baseTupleInRelsOid);
extern TSet generateAlternativeTupleSet(alg_fctx * fctx, TSet tset, RBITS relsOfTuples,
						 DTuple dtuple, TupleDesc tupleDesc, int tupleRelID);
extern HeapTuple clearValuesFromTSet(alg_fctx * fctx, TSet tset, char *nulls);

#endif   /* TSETFUNCS_H */
